/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___ShaderEffects___
#define ___ShaderEffects___

#include <LibGenerator/glew.h>
#include <wx/image.h>
#include <wx/glcanvas.h>
#include <wx/notebook.h>
#include <LibGenerator/ProceduralGenerator.h>
#include <LibGenerator/GlObjects.h>

namespace ProceduralTextures
{
	class ShaderEffects;
	class SePlugin;

	typedef enum
	{	eSEPARATION_NONE
	,	eSEPARATION_HORIZ
	,	eSEPARATION_VERT
	}	eSEPARATION;


	class Effect
	{
	private:
		wxImage *					m_pImage;
		std::vector <wxImage *>		m_arrayImages;
		std::vector <wxString>		m_arrayImagesFiles;
		GlShaderProgram			*	m_pShaderProgram;
		wxString					m_strVertexFile;
		wxString					m_strFragmentFile;
		bool						m_bNewShader;
		wxMilliClock_t				m_llStartTime;
		int							m_iWidth;
		int							m_iHeight;

	public:
		Effect( wxImage * p_pImage);
		virtual ~Effect();

		void SetVertexFile( const wxString & p_strPath);
		void SetFragmentFile( const wxString & p_strPath);
		void Compile();
		void SetImagePath( size_t p_iImage, const wxString & p_strImagePath);

		void GlPreRender();
		void GlPostRender();
		void GlCleanup();

		bool Activate( eSEPARATION p_eSeparationType, int p_iSeparationOffset);
		void Deactivate();

		wxString GetCompilerLog( eSHADER_OBJECT_TYPE p_eType);
		wxString GetLinkerLog();

		void UpdateImages();

		inline void ResetTime	()	{ m_llStartTime = wxGetLocalTimeMillis(); }
	};

	class ShaderEffects : public ProceduralGenerator
	{
	private:
		wxImage											m_image;
		std::vector<Effect *>							m_arrayEffects;
		eSEPARATION										m_eSeparationType;
		int												m_iSeparationOffset;
		GlFrameBuffer									m_frameBuffer1;
		GlFrameBuffer									m_frameBuffer2;
		GlTexture										m_texture1;
		GlTexture										m_texture2;
		Specifics< eCONTROL_TYPE_BUTTON >	m_specButtonImage;
		Specifics< eCONTROL_TYPE_STATIC >	m_specStaticSeparator;
		Specifics< eCONTROL_TYPE_COMBO >	m_specComboSeparator;
		Specifics< eCONTROL_TYPE_SLIDER >	m_specSliderOffset;
		Specifics< eCONTROL_TYPE_BUTTON >	m_specButtonReset;
		Specifics< eCONTROL_TYPE_STATIC >	m_specStaticShaders;
		Specifics< eCONTROL_TYPE_COMBO >	m_specComboShaders;
		Specifics< eCONTROL_TYPE_BUTTON >	m_specButtonVertexFile;
		Specifics< eCONTROL_TYPE_BUTTON >	m_specButtonFragmentFile;
		Specifics< eCONTROL_TYPE_BUTTON >	m_specButtonCompile;
		Specifics< eCONTROL_TYPE_BUTTON >	m_specButtonCompilerLog;
		Specifics< eCONTROL_TYPE_BUTTON >	m_specButtonRemove;
		Effect *										m_pSelectedEffect;

	public:
		ShaderEffects( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame);
		virtual ~ShaderEffects();

		void SetImagePath	( const wxString & p_strImagePath);
		void ResetTime		();

		bool		EffectAdd				();
		bool		EffectRemove			( size_t p_uiIndex);
		void		EffectSetVertexFile		( size_t p_uiIndex, const wxString & p_strPath);
		void		EffectSetFragmentFile	( size_t p_uiIndex, const wxString & p_strPath);
		void		EffectSetImagePath		( size_t p_uiIndex, size_t p_iImage, const wxString & p_strImagePath);
		void		EffectCompile			( size_t p_uiIndex);
		wxString	EffectGetCompilerLog	( size_t p_uiIndex, eSHADER_OBJECT_TYPE p_eType);
		wxString	EffectGetLinkerLog		( size_t p_uiIndex);
		bool		EffectActivate			( size_t p_uiIndex);
		void		EffectDeactivate		( size_t p_uiIndex);

		inline void SetSeparationOffset	( int p_iOffset)		{ m_iSeparationOffset = p_iOffset; }
		inline void SetSeparationType	( eSEPARATION p_eType)	{ m_eSeparationType = p_eType; }

	private:
		virtual void DoResize			( const wxSize & p_size);
		virtual void DoGlInit			();
		virtual void DoGlPreRender		();
		virtual void DoGlRender			( bool & p_bChanged);
		virtual void DoGlPostRender		();
		virtual void DoGlCleanup			();
		virtual void DoGeneratePanel	();

		typedef enum
		{	eID_IMAGEPATH		= 52
		,	eID_SEPTYPE			= 53
		,	eID_SEPOFFSET		= 54
		,	eID_RESETTIME		= 55
		,	eID_SHADERS			= 56
		,	eID_REMOVE			= 57
		,	eID_VERTEXFILE		= 58
		,	eID_FRAGMENTFILE	= 59
		,	eID_COMPILE			= 60
		,	eID_COMPILERLOG		= 61
		}	eID;

		void OnSepType				( wxCommandEvent	&	p_event );
		void OnSepOffset			( wxCommandEvent	&	p_event );
		void OnResetTime			( wxCommandEvent	&	p_event );
		void OnSelectShader			( wxCommandEvent	&	p_event );
		void OnRemove				( wxCommandEvent	&	p_event );
		void OnShaderCompile		( wxCommandEvent	&	p_event );
		void OnCompilerLog			( wxCommandEvent	&	p_event );
		void OnImagePath			( wxCommandEvent	&	p_event );
		void OnVertexShaderPath		( wxCommandEvent	&	p_event );
		void OnFragmentShaderPath	( wxCommandEvent	&	p_event );
	};

	class SePlugin : public Plugin< ShaderEffects, SePlugin >
	{
	public:
		SePlugin();
		virtual ~SePlugin();

		virtual wxString GetName();

		DECLARE_PLUGIN()
	};
}

#endif
