#include "BuggedLifeGame/GeneratorBuggedLifeGame.h"
#include <LibGenerator/Random.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************
/**
 *\~english
 *\brief		Function to substract 2 pixels given a minimal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_min	The minimal pixel
 *\return		(\p p_a - \p p_b) with each component of the result >= to the same component from \p p_min
 *\~french
 *\brief		Fonction pour soustraire 2 pixels avec un pixel minimum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_min	Le pixel minimal
 *\return		(\p p_a - \p p_b) avec chaque composant >=  la composante correspondante de \p p_min
 */
UbPixel Substract( const UbPixel & p_a, const UiPixel & p_b, const UbPixel & p_min )
{
	UbPixel l_result;

	if( p_a.r >= p_min.r + p_b.r )
	{
		l_result.r = p_a.r - p_b.r;
	}
	else
	{
		l_result.r = p_min.r;
	}

	if (p_a.g >= p_min.g + p_b.g)
	{
		l_result.g = p_a.g - p_b.g;
	}
	else
	{
		l_result.g = p_min.g;
	}

	if (p_a.b >= p_min.b + p_b.b)
	{
		l_result.b = p_a.b - p_b.b;
	}
	else
	{
		l_result.b = p_min.b;
	}

	if (p_a.a >= p_min.a + p_b.a)
	{
		l_result.a = p_a.a - p_b.a;
	}
	else
	{
		l_result.a = p_min.a;
	}

	return l_result;
}
/**
 *\~english
 *\brief		Function to add 2 pixels given a maximal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_max	The maximal pixel
 *\return		(\p p_a + \p p_b) with each component of the result <= to the same component from \p p_max
 *\~french
 *\brief		Fonction pour additionner 2 pixels avec un pixel maximum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_max	Le pixel maximal
 *\return		(\p p_a + \p p_b) avec chaque composante <=  la composante correspondante de \p p_max
 */
UbPixel Add( const UbPixel & p_a, const UiPixel & p_b, const UbPixel & p_max)
{
	UbPixel l_result;

	if (p_a.r + p_b.r < p_max.r)
	{
		l_result.r = p_a.r + p_b.r;
	}
	else
	{
		l_result.r = p_max.r;
	}

	if (p_a.g + p_b.g < p_max.g)
	{
		l_result.g = p_a.g + p_b.g;
	}
	else
	{
		l_result.g = p_max.g;
	}

	if (p_a.b + p_b.b < p_max.b)
	{
		l_result.b = p_a.b + p_b.b;
	}
	else
	{
		l_result.b = p_max.b;
	}

	if (p_a.a + p_b.a < p_max.a)
	{
		l_result.a = p_a.a + p_b.a;
	}
	else
	{
		l_result.a = p_max.a;
	}

	return l_result;
}

/**
 *\~english
 *\brief		Function to substract 2 pixels given a minimal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_min	The minimal pixel
 *\return		(\p p_a - \p p_b) with each component of the result >= to the same component from \p p_min
 *\~french
 *\brief		Fonction pour soustraire 2 pixels avec un pixel minimum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_min	Le pixel minimal
 *\return		(\p p_a - \p p_b) avec chaque composante >=  la composante correspondante de \p p_min
 */
UbPixel Substract( const UiPixel & p_a, const UiPixel & p_b, const UbPixel & p_min)
{
	UbPixel l_result;

	if (p_a.r >= p_min.r + p_b.r)
	{
		l_result.r = p_a.r - p_b.r;
	}
	else
	{
		l_result.r = p_min.r;
	}

	if (p_a.g >= p_min.g + p_b.g)
	{
		l_result.g = p_a.g - p_b.g;
	}
	else
	{
		l_result.g = p_min.g;
	}

	if (p_a.b >= p_min.b + p_b.b)
	{
		l_result.b = p_a.b - p_b.b;
	}
	else
	{
		l_result.b = p_min.b;
	}

	if (p_a.a >= p_min.a + p_b.a)
	{
		l_result.a = p_a.a - p_b.a;
	}
	else
	{
		l_result.a = p_min.a;
	}

	return l_result;
}
/**
 *\~english
 *\brief		Function to add 2 pixels given a maximal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_max	The maximal pixel
 *\return		(\p p_a + \p p_b) with each component of the result <= to the same component from \p p_max
 *\~french
 *\brief		Fonction pour additionner 2 pixels avec un pixel maximum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_max	Le pixel maximal
 *\return		(\p p_a + \p p_b) avec chaque composante <=  la composante correspondante de \p p_max
*/
UbPixel Add( const UiPixel & p_a, const UiPixel & p_b, const UbPixel & p_max)
{
	UbPixel l_result;

	if (p_a.r + p_b.r < p_max.r)
	{
		l_result.r = p_a.r + p_b.r;
	}
	else
	{
		l_result.r = p_max.r;
	}

	if (p_a.g + p_b.g < p_max.g)
	{
		l_result.g = p_a.g + p_b.g;
	}
	else
	{
		l_result.g = p_max.g;
	}

	if (p_a.b + p_b.b < p_max.b)
	{
		l_result.b = p_a.b + p_b.b;
	}
	else
	{
		l_result.b = p_max.b;
	}

	if (p_a.a + p_b.a < p_max.a)
	{
		l_result.a = p_a.a + p_b.a;
	}
	else
	{
		l_result.a = p_max.a;
	}

	return l_result;
}

//*************************************************************************************************

BlgCell :: BlgCell()
	:	m_alivePx( NULL)
	,	m_deadPx( NULL)
	,	m_nextPixel( NULL)
{
}

BlgCell :: BlgCell( const BlgCell & p_cell)
	:	m_alivePx( p_cell.m_alivePx)
	,	m_deadPx( p_cell.m_deadPx)
	,	m_nextPixel( p_cell.m_nextPixel)
	,	m_pixel( p_cell.m_pixel)
	,	m_step( p_cell.m_step)
	,	m_age( p_cell.m_age)
	,	m_med( p_cell.m_med)
{
	for (size_t i = 0 ; i < 8 ; i++)
	{
		m_neighbours[i] = p_cell.m_neighbours[i];
	}
}

BlgCell :: ~BlgCell()
{
}

BlgCell & BlgCell :: operator =( const BlgCell & p_cell)
{
	m_alivePx = p_cell.m_alivePx;
	m_deadPx = p_cell.m_deadPx;
	for (size_t i = 0 ; i < 8 ; i++)
	{
		m_neighbours[i] = p_cell.m_neighbours[i];
	}
	m_alive = p_cell.m_alive;
	m_nextPixel = p_cell.m_nextPixel;
	m_pixel = p_cell.m_pixel;
	m_step = p_cell.m_step;
	m_age = p_cell.m_age;
	m_med = p_cell.m_med;
	return * this;
}

void BlgCell :: Set( UbPixel * p_pixel, UbPixel * p_alivePx, UbPixel * p_deadPx, bool p_alive)
{
	m_alivePx = p_alivePx;
	m_deadPx = p_deadPx;
	m_nextPixel = p_pixel;

	if( m_alivePx && m_deadPx && m_nextPixel )
	{
		m_step.r = ((unsigned int)(m_alivePx->r) + 1 - m_deadPx->r) >> 5;
		m_step.g = ((unsigned int)(m_alivePx->g) + 1 - m_deadPx->g) >> 5;
		m_step.b = ((unsigned int)(m_alivePx->b) + 1 - m_deadPx->b) >> 5;
		m_step.a = ((unsigned int)(m_alivePx->a) + 1 - m_deadPx->a) >> 5;

		SetAlive( p_alive);
	}
}

void BlgCell :: Die()
{
	m_alive = false;

	*m_nextPixel = Substract( m_pixel, m_step, *m_deadPx);
	m_med = *m_nextPixel / m_step;
}

void BlgCell :: Live()
{
	if ( ! m_alive)
	{
		m_age.r = 0;
		m_age.g = 0;
		m_age.b = 0;
		m_age.a = 0;
	}

	m_alive = true;

	m_age += m_step;
	*m_nextPixel = Add( m_pixel, m_age, *m_alivePx);
	m_med = *m_nextPixel / m_step;
}

void BlgCell :: Stay()
{
	if (m_alive)
	{
		Live();
	}
	else
	{
		Die();
	}
}

void BlgCell :: SetAlive( bool p_alive)
{
	m_alive = p_alive;

	if (m_alive)
	{
		*m_nextPixel = *m_alivePx;
	}
	else
	{
		*m_nextPixel = *m_deadPx;
	}
}

void BlgCell :: Update()
{
	m_pixel.r = 0;
	m_pixel.g = 0;
	m_pixel.b = 0;
	m_pixel.a = 0;

	for (int i = 0 ; i < 8 ; i++)
	{
		m_pixel += m_neighbours[i]->m_med;
	}
}

//*************************************************************************************************

BuggedLifeGame::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	ProceduralGenerator::Thread( p_pParent, p_uiIndex, iWidth, iTop, iBottom, iTotalHeight, p_pxColour),
		m_cells( NULL)
{
}

BuggedLifeGame::Thread :: ~Thread()
{
}

void BuggedLifeGame::Thread :: Step()
{
	int l_nbAlive = 0;

	for( int i = 0 ; i < m_iWidth ; i++ )
	{
		for( int j = m_iTop ; j < m_iBottom ; j++ )
		{
			l_nbAlive = _getAliveNeightbours( i, j );

			if( l_nbAlive == 3 )
			{
				(*m_cells)[j][i].Live();
			}
			else if( l_nbAlive != 2 )
			{
				(*m_cells)[j][i].Die();
			}
			else
			{
				(*m_cells)[j][i].Stay();
			}
		}
	}
}

int BuggedLifeGame::Thread :: _getAliveNeightbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_iWidth || p_y < 0 || p_y >= m_iHeight)
	{
		return 0;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_iWidth - 1;
	}
	else if (p_x == m_iWidth - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_iHeight - 1;
	}
	else if (p_y == m_iHeight - 1)
	{
		l_down = 0;
	}

	return (	( *m_cells)[l_up	][l_left	].m_alive ? 1 : 0) + (	( *m_cells)[p_y		][l_left	].m_alive ? 1 : 0) + (	( *m_cells)[l_down	][l_left	].m_alive ? 1 : 0)
		   + (	( *m_cells)[l_up	][p_x		].m_alive ? 1 : 0) + (	( *m_cells)[l_down	][p_x		].m_alive ? 1 : 0)
		   + (	( *m_cells)[l_up	][l_right	].m_alive ? 1 : 0) + (	( *m_cells)[p_y		][l_right	].m_alive ? 1 : 0) + (	( *m_cells)[l_down	][l_right	].m_alive ? 1 : 0);
}

//*************************************************************************************************

BuggedLifeGame :: BuggedLifeGame( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator	( p_width, p_height, p_iWndId, p_pFrame										)
	,	m_cells				( Point<size_t, 2>( p_width, p_height )										)
	,	m_oldCells			( Point<size_t, 2>( p_width, p_height )										)
	,	m_specStatR			( Specifics< eCONTROL_TYPE_STATIC >( wxT( "R" ) )				)
	,	m_specStatG			( Specifics< eCONTROL_TYPE_STATIC >( wxT( "G" ) )				)
	,	m_specStatB			( Specifics< eCONTROL_TYPE_STATIC >( wxT( "B" ) )				)
	,	m_specSlideR		( Specifics< eCONTROL_TYPE_SLIDER >( 0, 255, m_pxColour.r )	)
	,	m_specSlideG		( Specifics< eCONTROL_TYPE_SLIDER >( 0, 255, m_pxColour.g )	)
	,	m_specSlideB		( Specifics< eCONTROL_TYPE_SLIDER >( 0, 255, m_pxColour.b )	)
{
	m_bNeedsConfigPanel = true;
	Random::srand ( time( NULL));

	m_alivePx = m_pxColour;

	m_deadPx.r = 0;
	m_deadPx.g = 0;
	m_deadPx.b = 0;
	m_deadPx.a = 0;

	bool l_alive;

	for (int i = 0 ; i < p_height ; i++)
	{
		for (int j = 0 ; j < p_width ; j++)
		{
			l_alive =  Random::rand() % 2 == 0;
			m_cells[i][j].Set( 		&m_frontBuffer[i][j], 	&m_alivePx, &m_deadPx, l_alive );
			m_oldCells[i][j].Set( 	&m_backBuffer[i][j], 	&m_alivePx, &m_deadPx, l_alive );
		}
	}

	for (int i = 0 ; i < p_width ; i++)
	{
		for (int j = 0 ; j < p_height ; j++)
		{
			DoInitialiseNeighbours( i, j);
		}
	}

	BuggedLifeGame::Thread * l_pThread;
	int l_iOffset = p_height / m_uiThreadCount;
	int l_iAdd;

	if( l_iOffset * m_uiThreadCount == p_height )
	{
		l_iAdd = 0;
	}
	else
	{
		l_iAdd = p_height - l_iOffset * m_uiThreadCount;
	}

	for( int i = 0 ; i < p_height ; i += l_iOffset )
	{
		l_pThread = CreateThread<BuggedLifeGame::Thread>( p_width, i, i + l_iOffset + l_iAdd, p_height, m_pxColour);
		l_pThread->Run();
		i += l_iAdd;
		l_iAdd = 0;
	}

	m_initialised = true;
	DoGeneratePanel();
}

BuggedLifeGame :: ~BuggedLifeGame()
{
	m_initialised = false;

	DoCleanup();

	m_cells.clear();
	m_oldCells.clear();
}

void BuggedLifeGame :: InitialiseStep()
{
	for (int i = 0 ; i < m_sizeImage.x ; i++)
	{
		for (int j = 0 ; j < m_sizeImage.y ; j++)
		{
			m_oldCells[j][i].Update();
		}
	}

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			static_cast<BuggedLifeGame::Thread *>( m_arraySlaveThreads[i])->SetBuffer( & m_oldCells);
		}
	}

	ProceduralGenerator::InitialiseStep();
}

void BuggedLifeGame :: DoInitialiseNeighbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_sizeImage.x || p_y < 0 || p_y >= m_sizeImage.y)
	{
		return;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_sizeImage.x-1;
	}
	else if (p_x == m_sizeImage.x - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_sizeImage.y-1;
	}
	else if (p_y == m_sizeImage.y - 1)
	{
		l_down = 0;
	}

	m_cells[p_y][p_x].m_neighbours[0] = & m_cells[l_up		][l_left	];
	m_cells[p_y][p_x].m_neighbours[1] = & m_cells[p_y		][l_left	];
	m_cells[p_y][p_x].m_neighbours[2] = & m_cells[l_down	][l_left	];
	m_cells[p_y][p_x].m_neighbours[3] = & m_cells[l_up		][p_x		];
	m_cells[p_y][p_x].m_neighbours[4] = & m_cells[l_down	][p_x		];
	m_cells[p_y][p_x].m_neighbours[5] = & m_cells[l_up		][l_right	];
	m_cells[p_y][p_x].m_neighbours[6] = & m_cells[p_y		][l_right	];
	m_cells[p_y][p_x].m_neighbours[7] = & m_cells[l_down	][l_right	];

	m_oldCells[p_y][p_x].m_neighbours[0] = & m_oldCells[l_up	][l_left	];
	m_oldCells[p_y][p_x].m_neighbours[1] = & m_oldCells[p_y		][l_left	];
	m_oldCells[p_y][p_x].m_neighbours[2] = & m_oldCells[l_down	][l_left	];
	m_oldCells[p_y][p_x].m_neighbours[3] = & m_oldCells[l_up	][p_x		];
	m_oldCells[p_y][p_x].m_neighbours[4] = & m_oldCells[l_down	][p_x		];
	m_oldCells[p_y][p_x].m_neighbours[5] = & m_oldCells[l_up	][l_right	];
	m_oldCells[p_y][p_x].m_neighbours[6] = & m_oldCells[p_y		][l_right	];
	m_oldCells[p_y][p_x].m_neighbours[7] = & m_oldCells[l_down	][l_right	];
}

int BuggedLifeGame :: _getAliveNeightbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_sizeImage.x || p_y < 0 || p_y >= m_sizeImage.y)
	{
		return 0;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_sizeImage.x - 1;
	}
	else if (p_x == m_sizeImage.x - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_sizeImage.y - 1;
	}
	else if (p_y == m_sizeImage.y - 1)
	{
		l_down = 0;
	}

	return (	m_cells[l_up	][l_left	].m_alive ? 1 : 0) + (	m_cells[p_y		][l_left	].m_alive ? 1 : 0) + (	m_cells[l_down	][l_left	].m_alive ? 1 : 0)
		   + (	m_cells[l_up	][p_x		].m_alive ? 1 : 0) + (	m_cells[l_down	][p_x		].m_alive ? 1 : 0)
		   + (	m_cells[l_up	][l_right	].m_alive ? 1 : 0) + (	m_cells[p_y		][l_right	].m_alive ? 1 : 0) + (	m_cells[l_down	][l_right	].m_alive ? 1 : 0);
}

void BuggedLifeGame :: DoGeneratePanel()
{
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_STATIC,	wxID_ANY,		wxPoint(  20,  0				), wxSize( 20, wxDEFAULT_HEIGHT	), &m_specStatR																												) );
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_SLIDER,	eRedValue,		wxPoint(  20, wxDEFAULT_HEIGHT	), wxSize( 20, 60				), &m_specSlideR,	wxSL_VERTICAL, eEVENT_TYPE_SLIDER_UPDATED, wxCommandEventHandler( BuggedLifeGame::OnSliderRed		)	) );

	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_STATIC,	wxID_ANY,		wxPoint(  60,  0				), wxSize( 60, wxDEFAULT_HEIGHT	), &m_specStatG																												) );
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_SLIDER,	eGreenValue,	wxPoint(  60, wxDEFAULT_HEIGHT	), wxSize( 20, 60				), &m_specSlideG,	wxSL_VERTICAL, eEVENT_TYPE_SLIDER_UPDATED, wxCommandEventHandler( BuggedLifeGame::OnSliderGreen	)	) );

	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_STATIC,	wxID_ANY,		wxPoint( 100,  0				), wxSize( 20, wxDEFAULT_HEIGHT	), &m_specStatB																												) );
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_SLIDER,	eBlueValue,		wxPoint( 100, wxDEFAULT_HEIGHT	), wxSize( 20, 60				), &m_specSlideB,	wxSL_VERTICAL, eEVENT_TYPE_SLIDER_UPDATED, wxCommandEventHandler( BuggedLifeGame::OnSliderBlue	)	) );
}

void BuggedLifeGame :: OnSliderRed( wxCommandEvent & p_event)
{
	SetRed( p_event.GetInt());
}

void BuggedLifeGame :: OnSliderGreen( wxCommandEvent & p_event)
{
	SetGreen( p_event.GetInt());
}

void BuggedLifeGame :: OnSliderBlue( wxCommandEvent & p_event)
{
	SetBlue( p_event.GetInt());
}

//******************************************************************************

IMPLEMENT_PLUGIN( BlgPlugin, BuggedLifeGame)

BlgPlugin :: BlgPlugin()
	:	Plugin< BuggedLifeGame, BlgPlugin >()
{
}

BlgPlugin :: ~BlgPlugin()
{
}

wxString BlgPlugin :: GetName()
{
#if defined( NDEBUG )
	return wxT( "Bugged Life Game");
#else
	return wxT( "Bugged Life Game (Debug)");
#endif
}

//*************************************************************************************************
