#include "ColoredLifeGame/GeneratorColoredLifeGame.h"
#include <LibGenerator/Random.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************
/**
 *\~english
 *\brief		Function to substract 2 pixels given a minimal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_min	The minimal pixel
 *\return		(\p p_a - \p p_b) with each component of the result >= to the same component from \p p_min
 *\~french
 *\brief		Fonction pour soustraire 2 pixels avec un pixel minimum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_min	Le pixel minimal
 *\return		(\p p_a - \p p_b) avec chaque composant >=  la composante correspondante de \p p_min
 */
UbPixel Substract( const UbPixel & p_a, const UiPixel & p_b, const UbPixel & p_min )
{
	UbPixel l_result;

	if( p_a.r >= p_min.r + p_b.r )
	{
		l_result.r = p_a.r - p_b.r;
	}
	else
	{
		l_result.r = p_min.r;
	}

	if (p_a.g >= p_min.g + p_b.g)
	{
		l_result.g = p_a.g - p_b.g;
	}
	else
	{
		l_result.g = p_min.g;
	}

	if (p_a.b >= p_min.b + p_b.b)
	{
		l_result.b = p_a.b - p_b.b;
	}
	else
	{
		l_result.b = p_min.b;
	}

	if (p_a.a >= p_min.a + p_b.a)
	{
		l_result.a = p_a.a - p_b.a;
	}
	else
	{
		l_result.a = p_min.a;
	}

	return l_result;
}
/**
 *\~english
 *\brief		Function to add 2 pixels given a maximal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_max	The maximal pixel
 *\return		(\p p_a + \p p_b) with each component of the result <= to the same component from \p p_max
 *\~french
 *\brief		Fonction pour additionner 2 pixels avec un pixel maximum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_max	Le pixel maximal
 *\return		(\p p_a + \p p_b) avec chaque composante <=  la composante correspondante de \p p_max
 */
UbPixel Add( const UbPixel & p_a, const UiPixel & p_b, const UbPixel & p_max)
{
	UbPixel l_result;

	if (p_a.r + p_b.r < p_max.r)
	{
		l_result.r = p_a.r + p_b.r;
	}
	else
	{
		l_result.r = p_max.r;
	}

	if (p_a.g + p_b.g < p_max.g)
	{
		l_result.g = p_a.g + p_b.g;
	}
	else
	{
		l_result.g = p_max.g;
	}

	if (p_a.b + p_b.b < p_max.b)
	{
		l_result.b = p_a.b + p_b.b;
	}
	else
	{
		l_result.b = p_max.b;
	}

	if (p_a.a + p_b.a < p_max.a)
	{
		l_result.a = p_a.a + p_b.a;
	}
	else
	{
		l_result.a = p_max.a;
	}

	return l_result;
}

/**
 *\~english
 *\brief		Function to substract 2 pixels given a minimal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_min	The minimal pixel
 *\return		(\p p_a - \p p_b) with each component of the result >= to the same component from \p p_min
 *\~french
 *\brief		Fonction pour soustraire 2 pixels avec un pixel minimum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_min	Le pixel minimal
 *\return		(\p p_a - \p p_b) avec chaque composante >=  la composante correspondante de \p p_min
 */
UbPixel Substract( const UiPixel & p_a, const UiPixel & p_b, const UbPixel & p_min)
{
	UbPixel l_result;

	if (p_a.r >= p_min.r + p_b.r)
	{
		l_result.r = p_a.r - p_b.r;
	}
	else
	{
		l_result.r = p_min.r;
	}

	if (p_a.g >= p_min.g + p_b.g)
	{
		l_result.g = p_a.g - p_b.g;
	}
	else
	{
		l_result.g = p_min.g;
	}

	if (p_a.b >= p_min.b + p_b.b)
	{
		l_result.b = p_a.b - p_b.b;
	}
	else
	{
		l_result.b = p_min.b;
	}

	if (p_a.a >= p_min.a + p_b.a)
	{
		l_result.a = p_a.a - p_b.a;
	}
	else
	{
		l_result.a = p_min.a;
	}

	return l_result;
}
/**
 *\~english
 *\brief		Function to add 2 pixels given a maximal pixel value for the result
 *\param[in]	p_a		The first pixel
 *\param[in]	p_b		The second pixel
 *\param[in]	p_max	The maximal pixel
 *\return		(\p p_a + \p p_b) with each component of the result <= to the same component from \p p_max
 *\~french
 *\brief		Fonction pour additionner 2 pixels avec un pixel maximum
 *\param[in]	p_a		Le premier pixel
 *\param[in]	p_b		Le second pixel
 *\param[in]	p_max	Le pixel maximal
 *\return		(\p p_a + \p p_b) avec chaque composante <=  la composante correspondante de \p p_max
*/
UbPixel Add( const UiPixel & p_a, const UiPixel & p_b, const UbPixel & p_max)
{
	UbPixel l_result;

	if (p_a.r + p_b.r < p_max.r)
	{
		l_result.r = p_a.r + p_b.r;
	}
	else
	{
		l_result.r = p_max.r;
	}

	if (p_a.g + p_b.g < p_max.g)
	{
		l_result.g = p_a.g + p_b.g;
	}
	else
	{
		l_result.g = p_max.g;
	}

	if (p_a.b + p_b.b < p_max.b)
	{
		l_result.b = p_a.b + p_b.b;
	}
	else
	{
		l_result.b = p_max.b;
	}

	if (p_a.a + p_b.a < p_max.a)
	{
		l_result.a = p_a.a + p_b.a;
	}
	else
	{
		l_result.a = p_max.a;
	}

	return l_result;
}

//*************************************************************************************************

ClgCell :: ClgCell()
	:	m_nextPixel		( NULL)
	,	m_deadPixel		( NULL)
	,	m_alivePixel	( NULL)
	,	m_stepPixel		( NULL)
{
}

ClgCell :: ClgCell( const ClgCell & cell)
	:	m_nextPixel		( cell.m_nextPixel)
	,	m_deadPixel		( cell.m_deadPixel)
	,	m_alivePixel	( cell.m_alivePixel)
	,	m_stepPixel		( cell.m_stepPixel)
	,	m_alive			( cell.m_alive)
{
}

ClgCell :: ~ClgCell()
{
}

ClgCell &  ClgCell :: operator =( const ClgCell & cell)
{
	m_nextPixel		= cell.m_nextPixel;
	m_deadPixel		= cell.m_deadPixel;
	m_alivePixel	= cell.m_alivePixel;
	m_stepPixel		= cell.m_stepPixel;
	m_alive			= cell.m_alive;

	return * this;
}

void ClgCell :: Set( UbPixel * p_pixel, UbPixel * p_deadPixel, UbPixel * p_alivePixel, UiPixel * p_stepPixel, bool p_alive)
{
	m_nextPixel		= p_pixel;
	m_deadPixel		= p_deadPixel;
	m_alivePixel	= p_alivePixel;
	m_stepPixel		= p_stepPixel;
	m_alive			= p_alive;

	if (m_alivePixel != NULL && m_deadPixel != NULL && m_nextPixel != NULL && m_stepPixel != NULL)
	{
		SetAlive( p_alive);

		if (m_alive)
		{
			m_nextPixel->r = m_alivePixel->r;
			m_nextPixel->g = m_alivePixel->g;
			m_nextPixel->b = m_alivePixel->b;
			m_nextPixel->a = m_alivePixel->a;
		}
		else
		{
			m_nextPixel->r = m_deadPixel->r;
			m_nextPixel->g = m_deadPixel->g;
			m_nextPixel->b = m_deadPixel->b;
			m_nextPixel->a = m_deadPixel->a;
		}
	}
}

void ClgCell :: Die()
{
	m_alive = false;
	*m_nextPixel = Add( *m_nextPixel, *m_stepPixel, *m_deadPixel);
}

void ClgCell :: Live()
{
	m_alive = true;

	m_nextPixel->r = m_alivePixel->r;
	m_nextPixel->g = m_alivePixel->g;
	m_nextPixel->b = m_alivePixel->b;
	m_nextPixel->a = m_alivePixel->a;
}

void ClgCell :: SetAlive( bool p_alive)
{
	bool l_alive = m_alive;

	if (p_alive)
	{
		Live();
	}
	else
	{
		Die();
	}
}

//*************************************************************************************************

ColoredLifeGame::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	ProceduralGenerator::Thread( p_pParent, p_uiIndex, iWidth, iTop, iBottom, iTotalHeight, p_pxColour)
	,	m_cells( NULL)
	,	m_oldCells( NULL)
{
}

ColoredLifeGame::Thread :: ~Thread()
{
}

void ColoredLifeGame::Thread :: Step()
{
	int l_iAlive;

	for (int i = 0 ; i < m_iWidth ; i++)
	{
		for (int j = m_iTop ; j < m_iBottom ; j++)
		{
			l_iAlive = _getAliveNeightbours( i, j);

			if (l_iAlive == 3)
			{
				(*m_cells)[j][i].Live();
			}
			else if (l_iAlive != 2)
			{
				(*m_cells)[j][i].Die();
			}
			else
			{
				(*m_cells)[j][i].SetAlive( (*m_oldCells)[j][i].m_alive);
			}
		}
	}
}

int ColoredLifeGame::Thread :: _getAliveNeightbours( int p_x, int p_y)
{
	if (p_x < 0 || p_x >= m_iWidth || p_y < 0 || p_y >= m_iHeight)
	{
		return 0;
	}

	int l_left = p_x - 1;
	int l_right = p_x + 1;
	int l_up = p_y - 1;
	int l_down = p_y + 1;

	if (p_x == 0)
	{
		l_left = m_iWidth-1;
	}
	else if (p_x == m_iWidth - 1)
	{
		l_right = 0;
	}

	if (p_y == 0)
	{
		l_up = m_iHeight-1;
	}
	else if (p_y == m_iHeight - 1)
	{
		l_down = 0;
	}

	int l_iResult = 0;

	if ((*m_oldCells)[l_up][l_left].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[p_y][l_left].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_down][l_left].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_up][p_x].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_down][p_x].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_up][l_right].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[p_y][l_right].m_alive)
	{
		l_iResult++;
	}

	if ((*m_oldCells)[l_down][l_right].m_alive)
	{
		l_iResult++;
	}

	return l_iResult;
}

//*************************************************************************************************

ColoredLifeGame :: ColoredLifeGame( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator			( p_width, p_height, p_iWndId, p_pFrame										)
	,	m_oldCells					( Point< int, 2 >( p_width, p_height )										)
	,	m_cells						( Point< int, 2 >( p_width, p_height )										)
	,	m_specStatR					( Specifics< eCONTROL_TYPE_STATIC >( wxT( "R" ) )				)
	,	m_specStatG					( Specifics< eCONTROL_TYPE_STATIC >( wxT( "G" ) )				)
	,	m_specStatB					( Specifics< eCONTROL_TYPE_STATIC >( wxT( "B" ) )				)
	,	m_specSlideR				( Specifics< eCONTROL_TYPE_SLIDER >( 0, 255, m_pxColour.r )	)
	,	m_specSlideG				( Specifics< eCONTROL_TYPE_SLIDER >( 0, 255, m_pxColour.g )	)
	,	m_specSlideB				( Specifics< eCONTROL_TYPE_SLIDER >( 0, 255, m_pxColour.b )	)
{
	m_bNeedsConfigPanel = true;
	m_pxColour.r = 0;
	m_pxColour.g = 0;
	m_pxColour.b = 0;
	Random::srand ( time( NULL));

	m_alivePixel = m_pxColour;

	m_deadPixel.r = 255;
	m_deadPixel.g = 255;
	m_deadPixel.b = 255;
	m_deadPixel.a = 255;

	m_stepPixel.r = abs( int(static_cast<unsigned int>( m_deadPixel.r) + 1 - static_cast<unsigned int>( m_alivePixel.r)) >> 4);
	m_stepPixel.g = abs( int(static_cast<unsigned int>( m_deadPixel.g) + 1 - static_cast<unsigned int>( m_alivePixel.g)) >> 4);
	m_stepPixel.b = abs( int(static_cast<unsigned int>( m_deadPixel.b) + 1 - static_cast<unsigned int>( m_alivePixel.b)) >> 4);
	m_stepPixel.a = abs( int(static_cast<unsigned int>( m_deadPixel.a) + 1 - static_cast<unsigned int>( m_alivePixel.a)) >> 4);

	bool l_alive;

	for (int i = 0 ; i < p_height ; i++)
	{
		for (int j = 0 ; j < p_width ; j++)
		{
			l_alive =  Random::rand() % 2 == 0;
			m_cells[i][j].Set( & m_frontBuffer[i][j], & m_deadPixel, & m_alivePixel, & m_stepPixel, l_alive);
			m_oldCells[i][j].Set( & m_backBuffer[i][j], & m_deadPixel, & m_alivePixel, & m_stepPixel, l_alive);
		}
	}

	ColoredLifeGame::Thread * l_pThread;
	int l_iOffset = p_height / m_uiThreadCount;
	int l_iAdd;

	if( l_iOffset * m_uiThreadCount == p_height )
	{
		l_iAdd = 0;
	}
	else
	{
		l_iAdd = p_height - l_iOffset * m_uiThreadCount;
	}

	for (int i = 0 ; i < p_height ; i += l_iOffset)
	{
		l_pThread = CreateThread<ColoredLifeGame::Thread>( p_width, i, i + l_iOffset + l_iAdd, p_height, m_pxColour);
		l_pThread->SetCells( & m_cells, & m_oldCells);
		l_pThread->Run();
		i += l_iAdd;
		l_iAdd = 0;
	}

	DoGeneratePanel();
	m_initialised = true;
}

ColoredLifeGame :: ~ColoredLifeGame()
{
	m_initialised = false;

	DoCleanup();

	m_cells.clear();
	m_oldCells.clear();
}

void ColoredLifeGame :: SwapBuffers()
{
	m_oldCells.swap( m_cells);
}

void ColoredLifeGame :: SetRed( unsigned char val)
{
	m_pxColour.r = val;
	m_alivePixel.r = val;

	m_stepPixel.r = abs( int(static_cast<unsigned int>( m_deadPixel.r) + 1 - static_cast<unsigned int>( m_alivePixel.r)) >> 4);
}
void ColoredLifeGame :: SetGreen( unsigned char val)
{
	m_pxColour.g = val;
	m_alivePixel.g = val;

	m_stepPixel.g = abs( int(static_cast<unsigned int>( m_deadPixel.g) + 1 - static_cast<unsigned int>( m_alivePixel.g)) >> 4);
}

void ColoredLifeGame :: SetBlue( unsigned char val)
{
	m_pxColour.b = val;
	m_alivePixel.b = val;

	m_stepPixel.b = abs( int(static_cast<unsigned int>( m_deadPixel.b) + 1 - static_cast<unsigned int>( m_alivePixel.b)) >> 4);
}

void ColoredLifeGame :: DoGeneratePanel()
{
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_STATIC,	wxID_ANY,		wxPoint(  20,  0				), wxSize( 20, wxDEFAULT_HEIGHT	), &m_specStatR																												) );
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_SLIDER,	eRedValue,		wxPoint(  20, wxDEFAULT_HEIGHT	), wxSize( 20, 60				), &m_specSlideR,	wxSL_VERTICAL, eEVENT_TYPE_SLIDER_UPDATED, wxCommandEventHandler(	ColoredLifeGame::OnSliderRed	)	) );

	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_STATIC,	wxID_ANY,		wxPoint(  60,  0				), wxSize( 60, wxDEFAULT_HEIGHT	), &m_specStatG																												) );
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_SLIDER,	eGreenValue,	wxPoint(  60, wxDEFAULT_HEIGHT	), wxSize( 20, 60				), &m_specSlideG,	wxSL_VERTICAL, eEVENT_TYPE_SLIDER_UPDATED, wxCommandEventHandler(	ColoredLifeGame::OnSliderGreen	)	) );

	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_STATIC,	wxID_ANY,		wxPoint( 100,  0				), wxSize( 20, wxDEFAULT_HEIGHT	), &m_specStatB																												) );
	m_arrayControls.push_back( ControlInfos( eCONTROL_TYPE_SLIDER,	eBlueValue,		wxPoint( 100, wxDEFAULT_HEIGHT	), wxSize( 20, 60				), &m_specSlideB,	wxSL_VERTICAL, eEVENT_TYPE_SLIDER_UPDATED, wxCommandEventHandler(	ColoredLifeGame::OnSliderBlue	)	) );
}

void ColoredLifeGame :: OnSliderRed( wxCommandEvent & p_event)
{
	SetRed( 255 - p_event.GetInt());
}

void ColoredLifeGame :: OnSliderGreen( wxCommandEvent & p_event)
{
	SetGreen( 255 - p_event.GetInt());
}

void ColoredLifeGame :: OnSliderBlue( wxCommandEvent & p_event)
{
	SetBlue( 255 - p_event.GetInt());
}

//******************************************************************************

IMPLEMENT_PLUGIN( ClgPlugin, ColoredLifeGame)

ClgPlugin :: ClgPlugin()
	:	Plugin< ColoredLifeGame, ClgPlugin >()
{
}

ClgPlugin :: ~ClgPlugin()
{
}

wxString ClgPlugin :: GetName()
{
#if defined( NDEBUG )
	return wxT( "Coloured Life Game");
#else
	return wxT( "Coloured Life Game (Debug)");
#endif
}

//*************************************************************************************************
