#include "LibGenerator/PrecompiledHeader.h"

#include "LibGenerator/ConfigPanel.h"
#include "LibGenerator/ProceduralGenerator.h"

using namespace ProceduralTextures;

//*************************************************************************************************

ControlInfos :: ControlInfos( eCONTROL_TYPE p_eType, wxWindowID p_iId, wxPoint const & p_ptPosition, wxSize const & p_szSize, SpecificsBase * p_pSpecific, int p_iStyle, eEVENT_TYPE p_eEventType, wxObjectEventFunction p_pfnFunction, bool p_bVisible )
	:	m_eType			( p_eType		)
	,	m_iId			( p_iId			)
	,	m_ptPosition	( p_ptPosition	)
	,	m_szSize		( p_szSize		)
	,	m_pSpecific		( p_pSpecific	)
	,	m_iStyle		( p_iStyle		)
	,	m_eEventType	( p_eEventType	)
	,	m_pfnFunction	( p_pfnFunction	)
	,	m_bVisible		( p_bVisible	)
{
}

ControlInfos :: ControlInfos( ControlInfos const & p_copy )
	:	m_eType			( p_copy.m_eType		)
	,	m_iId			( p_copy.m_iId			)
	,	m_ptPosition	( p_copy.m_ptPosition	)
	,	m_szSize		( p_copy.m_szSize		)
	,	m_pSpecific		( p_copy.m_pSpecific	)
	,	m_iStyle		( p_copy.m_iStyle		)
	,	m_eEventType	( p_copy.m_eEventType	)
	,	m_pfnFunction	( p_copy.m_pfnFunction	)
	,	m_bVisible		( p_copy.m_bVisible		)
{
}

ControlInfos :: ControlInfos( ControlInfos && p_copy )
	:	m_eType			()
	,	m_iId			()
	,	m_ptPosition	()
	,	m_szSize		()
	,	m_pSpecific		()
	,	m_iStyle		()
	,	m_eEventType	()
	,	m_pfnFunction	()
	,	m_bVisible		()
{
	m_eType			= p_copy.m_eType		;
	m_iId			= p_copy.m_iId			;
	m_ptPosition	= p_copy.m_ptPosition	;
	m_szSize		= p_copy.m_szSize		;
	m_pSpecific		= p_copy.m_pSpecific	;
	m_iStyle		= p_copy.m_iStyle		;
	m_eEventType	= p_copy.m_eEventType	;
	m_pfnFunction	= p_copy.m_pfnFunction	;
	m_bVisible		= p_copy.m_bVisible		;

	p_copy.m_eType			= eCONTROL_TYPE_NONE;
	p_copy.m_iId			= wxID_ANY;
	p_copy.m_ptPosition		= wxPoint();
	p_copy.m_szSize			= wxSize();
	p_copy.m_pSpecific		= NULL;
	p_copy.m_iStyle			= 0;
	p_copy.m_eEventType		= eEVENT_TYPE_NONE;
	p_copy.m_pfnFunction	= NULL;
	p_copy.m_bVisible		= false;
}

ControlInfos & ControlInfos :: operator =( ControlInfos const & p_copy )
{
	ControlInfos l_tmp( p_copy );
	std::swap( *this, l_tmp );
	return *this;
}

ControlInfos & ControlInfos :: operator =( ControlInfos && p_copy )
{
	ControlInfos l_tmp( p_copy );
	std::swap( *this, l_tmp );
	return *this;
}

ControlInfos :: ~ControlInfos()
{
}

//*************************************************************************************************

ConfigPanel :: ConfigPanel( wxEventType const (&p_arrayTypes)[eEVENT_TYPE_COUNT], ProceduralGenerator * p_pGenerator, wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:   wxPanel( p_pWindow, p_id, p_ptPosition, p_size)
	,	m_pGenerator( p_pGenerator )
{
	for( int i = 0 ; i < eEVENT_TYPE_COUNT ; i++ )
	{
		m_arrayTypes[i] = p_arrayTypes[i];
	}

	SetBackgroundColour( p_pWindow->GetBackgroundColour() );
	SetForegroundColour( p_pWindow->GetForegroundColour() );
	CreateSpecifics();
}

ConfigPanel :: ~ConfigPanel()
{
}

void ConfigPanel :: CreateSpecifics()
{
	std::for_each( m_pGenerator->BeginCtrls(), m_pGenerator->EndCtrls(), [&]( ControlInfos & p_ctrlInfos )
	{
		switch( p_ctrlInfos.m_eType )
		{
		case eCONTROL_TYPE_STATIC:	DoCreateStatic	( p_ctrlInfos );	break;
		case eCONTROL_TYPE_EDIT:	DoCreateEdit	( p_ctrlInfos );	break;
		case eCONTROL_TYPE_SLIDER:	DoCreateSlider	( p_ctrlInfos );	break;
		case eCONTROL_TYPE_COMBO:	DoCreateCombo	( p_ctrlInfos );	break;
		case eCONTROL_TYPE_BUTTON:	DoCreateButton	( p_ctrlInfos );	break;
		}
	});
}

void ConfigPanel :: DoCreateStatic( ControlInfos & p_ctrlInfos )
{
	wxString l_strText;
	Specifics< eCONTROL_TYPE_STATIC > * l_pSpec = NULL;

	if( p_ctrlInfos.m_pSpecific && p_ctrlInfos.m_pSpecific->m_eCtrlType == eCONTROL_TYPE_STATIC )
	{
		l_pSpec = static_cast< Specifics< eCONTROL_TYPE_STATIC > * >( p_ctrlInfos.m_pSpecific );
		l_strText = l_pSpec->m_strText;
		l_pSpec->m_pControl = new wxStaticText( this, p_ctrlInfos.m_iId, l_strText, p_ctrlInfos.m_ptPosition, p_ctrlInfos.m_szSize, p_ctrlInfos.m_iStyle );

		if( p_ctrlInfos.m_pfnFunction )
		{
			l_pSpec->m_pControl->Connect( p_ctrlInfos.m_iId, m_arrayTypes[p_ctrlInfos.m_eEventType], p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
		}

		if( ! p_ctrlInfos.m_bVisible )
		{
			l_pSpec->m_pControl->Hide();
		}
	}
}

void ConfigPanel :: DoCreateEdit( ControlInfos & p_ctrlInfos )
{
	wxString l_strText;
	Specifics< eCONTROL_TYPE_EDIT > * l_pSpec = NULL;

	if( p_ctrlInfos.m_pSpecific && p_ctrlInfos.m_pSpecific->m_eCtrlType == eCONTROL_TYPE_EDIT )
	{
		l_pSpec = static_cast< Specifics< eCONTROL_TYPE_EDIT > * >( p_ctrlInfos.m_pSpecific );
		l_strText = l_pSpec->m_strText;
		l_pSpec->m_pControl = new wxTextCtrl( this, p_ctrlInfos.m_iId, l_strText, p_ctrlInfos.m_ptPosition, p_ctrlInfos.m_szSize, p_ctrlInfos.m_iStyle );

		if( p_ctrlInfos.m_pfnFunction )
		{
			l_pSpec->m_pControl->Connect( p_ctrlInfos.m_iId, m_arrayTypes[p_ctrlInfos.m_eEventType], p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
		}

		if( ! p_ctrlInfos.m_bVisible )
		{
			l_pSpec->m_pControl->Hide();
		}
	}
}

void ConfigPanel :: DoCreateSlider( ControlInfos & p_ctrlInfos )
{
	int l_iMin = 0;
	int l_iMax = 100;
	int l_iVal = 0;
	Specifics< eCONTROL_TYPE_SLIDER > * l_pSpec = NULL;

	if( p_ctrlInfos.m_pSpecific && p_ctrlInfos.m_pSpecific->m_eCtrlType == eCONTROL_TYPE_SLIDER )
	{
		l_pSpec = static_cast< Specifics< eCONTROL_TYPE_SLIDER > * >( p_ctrlInfos.m_pSpecific );
		l_iMin = l_pSpec->m_iMin;
		l_iMax = l_pSpec->m_iMax;
		l_iVal = l_pSpec->m_iValue;
		l_pSpec->m_pControl = new wxSlider( this, p_ctrlInfos.m_iId, l_iVal, l_iMin, l_iMax, p_ctrlInfos.m_ptPosition, p_ctrlInfos.m_szSize, p_ctrlInfos.m_iStyle );

		if( p_ctrlInfos.m_pfnFunction )
		{
			l_pSpec->m_pControl->Connect( p_ctrlInfos.m_iId, m_arrayTypes[p_ctrlInfos.m_eEventType], p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
		}

		if( ! p_ctrlInfos.m_bVisible )
		{
			l_pSpec->m_pControl->Hide();
		}
	}
}

void ConfigPanel :: DoCreateCombo( ControlInfos & p_ctrlInfos )
{
	wxArrayString l_arrayValues;
	wxString l_strValue;
	Specifics< eCONTROL_TYPE_COMBO > * l_pSpec = NULL;

	if( p_ctrlInfos.m_pSpecific && p_ctrlInfos.m_pSpecific->m_eCtrlType == eCONTROL_TYPE_COMBO )
	{
		l_pSpec = static_cast< Specifics< eCONTROL_TYPE_COMBO > * >( p_ctrlInfos.m_pSpecific );
		l_arrayValues = l_pSpec->m_arrayValues;
		l_strValue = l_pSpec->m_strValue;
		l_pSpec->m_pControl = new wxComboBox( this, p_ctrlInfos.m_iId, l_strValue, p_ctrlInfos.m_ptPosition, p_ctrlInfos.m_szSize, l_arrayValues, p_ctrlInfos.m_iStyle );

		if( p_ctrlInfos.m_pfnFunction )
		{
			l_pSpec->m_pControl->Connect( p_ctrlInfos.m_iId, m_arrayTypes[p_ctrlInfos.m_eEventType], p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
		}

		if( ! p_ctrlInfos.m_bVisible )
		{
			l_pSpec->m_pControl->Hide();
		}
	}
}

void ConfigPanel :: DoCreateButton( ControlInfos & p_ctrlInfos )
{
	wxString l_strText;
	Specifics< eCONTROL_TYPE_BUTTON > * l_pSpec = NULL;

	if( p_ctrlInfos.m_pSpecific && p_ctrlInfos.m_pSpecific->m_eCtrlType == eCONTROL_TYPE_BUTTON )
	{
		l_pSpec = static_cast< Specifics< eCONTROL_TYPE_BUTTON > * >( p_ctrlInfos.m_pSpecific );
		l_strText = l_pSpec->m_strText;
		l_pSpec->m_pControl = new wxButton( this, p_ctrlInfos.m_iId, l_strText, p_ctrlInfos.m_ptPosition, p_ctrlInfos.m_szSize, p_ctrlInfos.m_iStyle );

		if( p_ctrlInfos.m_pfnFunction )
		{
			if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_TEXT_FILEA )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadTextFileA ) );
				m_pGenerator->Connect( eEVENT_TEXTA, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_TEXT_FILEB )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadTextFileB ) );
				m_pGenerator->Connect( eEVENT_TEXTB, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_IMAGEA )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadImageA ) );
				m_pGenerator->Connect( eEVENT_IMAGEA, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_IMAGEB )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadImageB ) );
				m_pGenerator->Connect( eEVENT_IMAGEB, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_BINARY_FILEA )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadBinaryFileA ) );
				m_pGenerator->Connect( eEVENT_BINARYA, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_BINARY_FILEB )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadBinaryFileB ) );
				m_pGenerator->Connect( eEVENT_BINARYB, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_COLOURA )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadColourA ) );
				m_pGenerator->Connect( eEVENT_COLOURA, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else if( p_ctrlInfos.m_eEventType == eEVENT_TYPE_LOAD_COLOURB )
			{
				Connect( p_ctrlInfos.m_iId, m_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED], wxCommandEventHandler( ConfigPanel::OnLoadColourB ) );
				m_pGenerator->Connect( eEVENT_COLOURB, wxEVT_NULL, p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
			else
			{
				l_pSpec->m_pControl->Connect( p_ctrlInfos.m_iId, m_arrayTypes[p_ctrlInfos.m_eEventType], p_ctrlInfos.m_pfnFunction, NULL, m_pGenerator );
			}
		}

		if( ! p_ctrlInfos.m_bVisible )
		{
			l_pSpec->m_pControl->Hide();
		}
	}
}

void ConfigPanel :: OnLoadImageA( wxCommandEvent & p_event )
{
	wxString l_strWildcard = wxT( "All supported files (*.bmp;*.gif;*.png;*.jpg)|*.bmp;*.gif;*.png;*.jpg|BITMAP files (*.bmp)|*.bmp|GIF files(*.gif)|*.gif|JPEG files (*.jpg)|*.jpg|PNG files (*.png)|*.png");
	wxString l_strImagePath;
	wxFileDialog l_dialog( this, wxT( "Choose a picture" ), wxT( "./data" ), wxEmptyString, l_strWildcard );

	if( l_dialog.ShowModal() == wxID_OK )
	{
		l_strImagePath = l_dialog.GetPath();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_IMAGEA );
		l_event.SetString( l_strImagePath );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

void ConfigPanel :: OnLoadImageB( wxCommandEvent & p_event )
{
	wxString l_strWildcard = wxT( "All supported files (*.bmp;*.gif;*.png;*.jpg)|*.bmp;*.gif;*.png;*.jpg|BITMAP files (*.bmp)|*.bmp|GIF files(*.gif)|*.gif|JPEG files (*.jpg)|*.jpg|PNG files (*.png)|*.png");
	wxString l_strImagePath;
	wxFileDialog l_dialog( this, wxT( "Choose a picture" ), wxT( "./data" ), wxEmptyString, l_strWildcard );

	if( l_dialog.ShowModal() == wxID_OK )
	{
		l_strImagePath = l_dialog.GetPath();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_IMAGEB );
		l_event.SetString( l_strImagePath );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

void ConfigPanel :: OnLoadTextFileA( wxCommandEvent & p_event )
{
	wxString l_strFile;
	wxString l_strWildcard = wxT( "All files (*.*)|*.*");
	wxFileDialog l_dialog( this, wxT( "Choose a file" ), wxT( "./data" ), wxEmptyString, l_strWildcard );

	if( l_dialog.ShowModal() == wxID_OK )
	{
		l_strFile = l_dialog.GetPath();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_TEXTA );
		l_event.SetString( l_strFile );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

void ConfigPanel :: OnLoadTextFileB( wxCommandEvent & p_event )
{
	wxString l_strFile;
	wxString l_strWildcard = wxT( "All files (*.*)|*.*");
	wxFileDialog l_dialog( this, wxT( "Choose a file" ), wxT( "./data" ), wxEmptyString, l_strWildcard );

	if( l_dialog.ShowModal() == wxID_OK )
	{
		l_strFile = l_dialog.GetPath();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_TEXTB );
		l_event.SetString( l_strFile );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

void ConfigPanel :: OnLoadBinaryFileA( wxCommandEvent & p_event )
{
	wxString l_strFile;
	wxString l_strWildcard = wxT( "All files (*.*)|*.*");
	wxFileDialog l_dialog( this, wxT( "Choose a file" ), wxT( "./data" ), wxEmptyString, l_strWildcard );

	if( l_dialog.ShowModal() == wxID_OK )
	{
		l_strFile = l_dialog.GetPath();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_BINARYA );
		l_event.SetString( l_strFile );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

void ConfigPanel :: OnLoadBinaryFileB( wxCommandEvent & p_event )
{
	wxString l_strFile;
	wxString l_strWildcard = wxT( "All files (*.*)|*.*");
	wxFileDialog l_dialog( this, wxT( "Choose a file" ), wxT( "./data" ), wxEmptyString, l_strWildcard );

	if( l_dialog.ShowModal() == wxID_OK )
	{
		l_strFile = l_dialog.GetPath();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_BINARYB );
		l_event.SetString( l_strFile );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

void ConfigPanel :: OnLoadColourA( wxCommandEvent & p_event )
{
	wxColourDialog l_dialog( NULL );

	if (l_dialog.ShowModal() == wxID_OK)
	{
		wxColour l_colour = l_dialog.GetColourData().GetColour();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_COLOURA );
		l_event.SetInt( l_colour.GetPixel() );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

void ConfigPanel :: OnLoadColourB( wxCommandEvent & p_event )
{
	wxColourDialog l_dialog( NULL );

	if (l_dialog.ShowModal() == wxID_OK)
	{
		wxColour l_colour = l_dialog.GetColourData().GetColour();
		wxCommandEvent l_event( wxEVT_NULL, eEVENT_COLOURB );
		l_event.SetInt( l_colour.GetPixel() );
		m_pGenerator->AddPendingEvent( l_event );
	}
}

//*************************************************************************************************
