#include "ProceduralGenerator/PrecompiledHeader.h"

#include "ProceduralGenerator/MainFrame.h"
#include "ProceduralGenerator/Application.h"
#include "ProceduralGenerator/ProjectListDialog.h"
#include "ProceduralGenerator/RenderPanel.h"
#include "ProceduralGenerator/StatusBar.h"
#include "../resource/ProceduralGenerator/exit.xpm"
#include "../resource/ProceduralGenerator/print_screen.xpm"
#include "../resource/ProceduralGenerator/select.xpm"

#include <wx/event.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

IMPLEMENT_CLASS( MainFrame, wxFrame);

MainFrame :: MainFrame()
	:	wxFrame( NULL, eMyID, wxT( "Procedural Generator"), wxPoint( 0, 0), wxDefaultSize, wxRESIZE_BORDER | wxMAXIMIZE_BOX | wxMINIMIZE_BOX | wxSYSTEM_MENU | wxCAPTION | wxCLOSE_BOX | wxCLIP_CHILDREN)
	,	m_pRenderPanel	( NULL					)
	,	m_pConfigPanel	( NULL					)
	,	m_pPlugin		( NULL					)
	,	m_pStatusBar	( NULL					)
{
	SetBackgroundColour( * wxBLACK);
	SetForegroundColour( * wxWHITE);
	m_pStatusBar = new StatusBar( this);
	SetStatusBar( m_pStatusBar);
	Connect( eEVENT_CPU_STEP_END, wxEVT_NULL, wxCommandEventHandler( MainFrame::OnCpuStepEnd ) );
}

MainFrame :: ~MainFrame()
{
}

void MainFrame :: Initialise()
{
	wxSize l_size( 256, 256);
	SetClientSize( l_size);
	DoLoadPlugins();
	DoBuildMenuBar();
	int l_attribList[] = { WX_GL_RGBA, WX_GL_LEVEL, 0, WX_GL_DOUBLEBUFFER };
	m_pRenderPanel = new RenderPanel( this, l_attribList, wxPoint( 0, 0), l_size);
	m_pRenderPanel->SetBackgroundColour( GetBackgroundColour());
	m_pRenderPanel->SetForegroundColour( GetForegroundColour());
	Show();
}

void MainFrame :: DoBuildMenuBar()
{
	wxPNGHandler * l_pHandler = new wxPNGHandler;
	wxImage::AddHandler( l_pHandler);
	wxToolBar * l_pToolbar = CreateToolBar( wxTB_FLAT | wxTB_HORIZONTAL);
	wxMemoryInputStream l_isSelect( select_png, sizeof( select_png));
	wxImage l_imgSelect( l_isSelect, wxBITMAP_TYPE_PNG);
	wxMemoryInputStream l_isPrint( print_screen_png, sizeof( print_screen_png));
	wxImage l_imgPrint( l_isPrint, wxBITMAP_TYPE_PNG);
	wxMemoryInputStream l_isExit( exit_png, sizeof( exit_png));
	wxImage l_imgExit( l_isExit, wxBITMAP_TYPE_PNG);
	l_pToolbar->SetBackgroundColour( GetBackgroundColour());
	l_pToolbar->SetForegroundColour( GetForegroundColour());
	l_pToolbar->AddTool( eSelectGenerator,	wxT( "Select Generator"),	l_imgSelect.Rescale( 32, 32, wxIMAGE_QUALITY_HIGH));
	l_pToolbar->AddTool( ePrintScreen,		wxT( "Print screen"),		l_imgPrint.Rescale( 32, 32, wxIMAGE_QUALITY_HIGH));
	l_pToolbar->AddTool( eExit,				wxT( "Exit"),				l_imgExit.Rescale( 32, 32, wxIMAGE_QUALITY_HIGH));
//	wxImage::RemoveHandler( l_pHandler->GetName());
	l_pToolbar->Realize();
}

ProceduralGenerator * MainFrame :: DoSelectGenerator()
{
	ProceduralGenerator * l_pReturn = NULL;
	wxArrayString l_choices;

	for (std::map<wxString, PluginBase *>::iterator l_it = m_mapPluginsByName.begin() ; l_it != m_mapPluginsByName.end() ; ++l_it)
	{
		l_choices.push_back( l_it->first);
	}

	ProjectListDialog l_dialog( this, l_choices);

	if (l_dialog.ShowModal() == wxID_OK)
	{
		wxString l_strProjectName = l_dialog.GetProject();
		l_pReturn = DoLoadGenerator( l_strProjectName, l_dialog.GetResolution());
	}

	return l_pReturn;
}

ProceduralGenerator * MainFrame :: DoLoadGenerator( const wxString & p_strName, const wxSize & p_size)
{
	ProceduralGenerator * l_pReturn = NULL;
	std::map<wxString, PluginBase *>::iterator l_it = m_mapPluginsByName.find( p_strName);

	if (l_it != m_mapPluginsByName.end())
	{
		m_pPlugin = l_it->second;
		SetClientSize( p_size);
		l_pReturn = l_it->second->CreateGenerator( p_size.x, p_size.y, MainFrame::eMyID, this);

		if (l_pReturn->NeedsConfigPanel())
		{
			wxEventType l_arrayTypes[eEVENT_TYPE_COUNT];
			l_arrayTypes[eEVENT_TYPE_SLIDER_UPDATED]		= wxEVT_COMMAND_SLIDER_UPDATED;
			l_arrayTypes[eEVENT_TYPE_SCROLL_THUMBRELEASE]	= wxEVT_SCROLL_THUMBRELEASE;
			l_arrayTypes[eEVENT_TYPE_SCROLL_THUMBTRACK]		= wxEVT_SCROLL_THUMBTRACK;
			l_arrayTypes[eEVENT_TYPE_COMBOBOX_SELECTED]		= wxEVT_COMMAND_COMBOBOX_SELECTED;
			l_arrayTypes[eEVENT_TYPE_BUTTON_CLIKED]			= wxEVT_COMMAND_BUTTON_CLICKED;
			l_arrayTypes[eEVENT_TYPE_TEXT_UPDATED]			= wxEVT_COMMAND_TEXT_UPDATED;
			l_arrayTypes[eEVENT_TYPE_TEXT_ENTER]			= wxEVT_COMMAND_TEXT_ENTER;
			SetClientSize( p_size.x + 150, p_size.y);
			m_pConfigPanel = new ConfigPanel( l_arrayTypes, l_pReturn, this, wxID_ANY, wxPoint( p_size.x, 0 ), wxSize( 150, p_size.y ) );
		}

		SetLabel( p_strName);
	}

	return l_pReturn;
}

void MainFrame :: DoDestroyGenerator()
{
	if( m_pConfigPanel )
	{
		m_pConfigPanel->Destroy();
		m_pConfigPanel = NULL;
	}
}

void MainFrame :: DoLoadPlugins()
{
	wxStandardPathsBase & l_paths = wxStandardPaths::Get();
	wxFileName l_dataDir = l_paths.GetExecutablePath();
	wxArrayString l_arrayDirs = l_dataDir.GetDirs();
	wxString l_strPluginsPath = l_dataDir.GetVolume() + l_dataDir.GetVolumeSeparator();

	for( size_t i = 0 ; i < l_arrayDirs.size() && l_arrayDirs[i] != wxT( "bin" ) ; i++ )
	{
		l_strPluginsPath += wxFileName::GetPathSeparator() + l_arrayDirs[i];
	}

	l_strPluginsPath = l_strPluginsPath + wxFileName::GetPathSeparator() + wxT( "share" ) + wxFileName::GetPathSeparator() + wxT( "ProceduralGenerator" ) + wxFileName::GetPathSeparator();

	if( l_strPluginsPath.find( wxT( "build_temp" ) ) != wxString::npos )
	{
		l_strPluginsPath.Replace( wxString( "share" ) + wxFileName::GetPathSeparator() + wxT( "ProceduralGenerator" ), wxT( "plugins" ) );
#if defined( NDEBUG )
		l_strPluginsPath += wxString( "Release" ) + wxFileName::GetPathSeparator();
#else
		l_strPluginsPath += wxString( "Debug" ) + wxFileName::GetPathSeparator();
#endif
	}

	DoLoadPlugins( l_strPluginsPath );
}

void MainFrame :: DoLoadPlugins( wxString const & p_strPath )
{
//	wxMessageBox( wxT( "DoLoadPlugins :\n" ) + p_strPath );
	wxDir l_dir( p_strPath );
	wxString l_strFileName;

	if (l_dir.IsOpened())
	{
		bool l_bRes = l_dir.GetFirst( & l_strFileName, wxString( wxT( "*" ) ) + wxDynamicLibrary::GetDllExt(), wxDIR_FILES );

		while (l_bRes)
		{
			wxDynamicLibrary * l_pDll = new wxDynamicLibrary;

			if( l_pDll->Load( p_strPath + l_strFileName ) )
			{
				Sleep( 100 );

				if (l_pDll->HasSymbol( wxT( "CreatePlugin" ) ) )
				{
					wxDYNLIB_FUNCTION( PCreatePluginFunction, CreatePlugin, *l_pDll );

					if (pfnCreatePlugin != NULL)
					{
						PluginBase * l_pPlugin = pfnCreatePlugin( l_pDll, wxTheApp );

						if (l_pPlugin != NULL)
						{
							m_mapPluginsByName.insert( std::pair<wxString, PluginBase *>( l_pPlugin->GetName(), l_pPlugin ) );
						}
					}
				}
				else
				{
					delete l_pDll;
				}
			}
			else
			{
				delete l_pDll;
			}

			l_bRes = l_dir.GetNext( & l_strFileName);
		}
	}
}

void MainFrame :: DoClose()
{
	Hide();

	if (m_pRenderPanel != NULL)
	{
		m_pRenderPanel->SetGenerator( NULL );
		m_pRenderPanel->Close();
		delete m_pRenderPanel;
		m_pRenderPanel = NULL;
	}

	DoDestroyGenerator();

	for (std::map<wxString, PluginBase *>::iterator l_it = m_mapPluginsByName.begin() ; l_it != m_mapPluginsByName.end() ; ++l_it)
	{
		l_it->second->Destroy();
	}

	m_mapPluginsByName.clear();
}

BEGIN_EVENT_TABLE( MainFrame, wxFrame)
	EVT_PAINT(											MainFrame::OnPaint				)
	EVT_SIZE(											MainFrame::OnSize				)
	EVT_CLOSE(											MainFrame::OnClose				)
	EVT_KEY_UP(											MainFrame::OnKeyUp				)
	EVT_MENU( 			eExit,							MainFrame::OnExit				)
	EVT_MENU( 			eSelectGenerator,				MainFrame::OnSelectGenerator	)
	EVT_MENU( 			ePrintScreen,					MainFrame::OnPrintScreen		)
END_EVENT_TABLE()

void MainFrame :: OnPaint( wxPaintEvent & p_event )
{
	wxPaintDC l_dc( this );
	p_event.Skip();
}

void MainFrame :: OnSize( wxSizeEvent & p_event)
{
	wxSize sizeClient = GetClientSize();

	if (m_pConfigPanel != NULL)
	{
		sizeClient.x -= m_pConfigPanel->GetClientSize().x;
	}

	if (m_pRenderPanel != NULL)
	{
		m_pRenderPanel->SetSize( 0, 0, sizeClient.x, sizeClient.y, wxSIZE_FORCE);
	}

	if (m_pConfigPanel != NULL)
	{
		m_pConfigPanel->SetSize( sizeClient.x, 0, 150, sizeClient.y, wxSIZE_FORCE);
	}

	p_event.Skip();
}

void MainFrame :: OnClose( wxCloseEvent & p_event)
{
	DoClose();
	p_event.Skip();
}

void MainFrame :: OnKeyUp( wxKeyEvent & p_event)
{
	if (m_pRenderPanel != NULL && p_event.GetKeyCode() == WXK_F12)
	{
		m_pRenderPanel->SaveFrame();
	}

	p_event.Skip();
}

void MainFrame :: OnExit( wxCommandEvent & p_event)
{
	Close( true );
	p_event.Skip();
}

void MainFrame :: OnSelectGenerator( wxCommandEvent & p_event)
{
	m_pRenderPanel->SetGenerator( NULL );
	DoDestroyGenerator();
	ProceduralGenerator * l_pGenerator = DoSelectGenerator();
	Maximize( false);
	m_pRenderPanel->SetGenerator( l_pGenerator);
	m_savedTime = wxGetLocalTimeMillis();
	p_event.Skip();
}

void MainFrame :: OnPrintScreen( wxCommandEvent & p_event)
{
	if (m_pRenderPanel != NULL)
	{
		m_pRenderPanel->SaveFrame();
	}

	p_event.Skip();
}

void MainFrame :: OnCpuStepEnd( wxCommandEvent & p_event )
{
	m_time = wxGetLocalTimeMillis();
	wxString l_strStatus = wxT( "C : ");
	wxMilliClock_t l_diff = m_time - m_savedTime;
	m_savedTime = wxGetLocalTimeMillis();

	if( l_diff != 0 )
	{
		std::basic_stringstream<wxChar> l_streamMs;
		std::basic_stringstream<wxChar> l_streamFps;
		l_streamMs.width( 5);
		l_streamMs.fill( wxT( '0'));
		l_streamMs << std::right << l_diff.GetValue();
		l_streamFps.width( 4);
		l_streamFps.fill( wxT( '0'));
		l_streamFps << std::right << (1000 / l_diff.GetValue());
		m_pStatusBar->SetStatusText( l_strStatus + l_streamMs.str() + wxT( "ms - ") + l_streamFps.str() + wxT( " FPS"), 0 );
		m_pStatusBar->Refresh();
	}

	if( m_pRenderPanel )
	{
		m_pRenderPanel->Render();

		m_time = wxGetLocalTimeMillis();
		l_strStatus = wxT( "R : ");
		l_diff = m_time - m_savedTime;
		m_savedTime = wxGetLocalTimeMillis();

		if( l_diff != 0 )
		{
			std::basic_stringstream<wxChar> l_streamMs;
			std::basic_stringstream<wxChar> l_streamFps;
			l_streamMs.width( 5);
			l_streamMs.fill( wxT( '0'));
			l_streamMs << std::right << l_diff.GetValue();
			l_streamFps.width( 4);
			l_streamFps.fill( wxT( '0'));
			l_streamFps << std::right << (1000 / l_diff.GetValue());
			m_pStatusBar->SetStatusText( l_strStatus + l_streamMs.str() + wxT( "ms - ") + l_streamFps.str() + wxT( " FPS"), 1 );
			m_pStatusBar->Refresh();
		}
	}

	p_event.Skip();
}

//*************************************************************************************************
